/*************************************************************************

		multi_tsp.c, v 1.01  90/11/07  13:25:43   author CH

**************************************************************************/

/*
Server fuer die Parallelisierung mit einer Markov-Kette
*/

#include <stdio.h> 
#include <signal.h>

#include "multi_tsp.h"
#include "haeufig.h"

extern	double	atof();
extern	void	exit();

#define	ALARM	5			/* 5 Minuten alive-Test */

/*-------------------------------------------*/
/* Variablen */

int	LinkId;

int	n;
int	anzRandomKonf = 10000;					/* Anzahl von random Konfigurationen */
int	anz_Laeufe = 20;					/* Anzahl der Laeufe */
int	anz_Proz = -1;						/* Anzahl der Prozessoren */

char	occamName[200] = "occam_src/xx";
char	instanzName[200] = "instanzen/tsp442.instanz";

int	debugflag = FALSE;				/* default: Keine Debugausgaben */
int	lockFileFlag = TRUE;				/* default: Lockfile wird benutzt */

int	alive = 0;						/* Occamprogramm lebt noch */

HaeufigkeitRec	g_zeit;
HaeufigkeitRec	g_wert;
HaeufigkeitRec	g_schritte;
HaeufigkeitRec	g_akzepts;
HaeufigkeitRec	g_nonakzepts;

/*------------------------------------------------------------------------------------------------------------*/

void	Error( str )
char	*str;
{
	fprintf( stderr, "%s\n", str );
}

void usage( name )
char	*name;
{
	fprintf( stderr,
		"usage: %s	linkname <occambin> <instanz> #Konfigurationen #Laeufe [-l][-d]\n", name );
	exit( 1 );
}

/*---------------------------------------------------------------------*/

int	OccamVersion()
{
#define	VERSIONNR	-9

	int	version = VERSIONNR;


	if ( SendInt( LinkId, 50, &version ) != 0 ) {
		Error( "Fehler: OCCAM Programm empfaengt keine Versionsnummer\n");
		return -1;
	}
	if ( ReceiveInt( LinkId, 50, &version ) != 0 ) {
		Error( "Fehler: OCCAM Programm sendet keine Versionsnummer\n");
		return -1;
	}
	if ( version != VERSIONNR ) {
		Error( "Fehler: OCCAM Programm sendet FALSCHE Versionsnummer\n" );
		return -1;
	}
	if ( ReceiveInt( LinkId, 50, &anz_Proz ) != 0 ) {
		Error( "Fehler: OCCAM Programm sendet keine Prozessoranzahl\n");
		return -1;
	}
	return 0;
}

/*---------------------------------------------------------------------*/

void	SetzeLink()
{
	int	err;
	char	ErrMsg[100];

	err = LinkAlloc( LinkId, ErrMsg, lockFileFlag );
	if ( err != LinkId ) {
		Error( ErrMsg );
		exit(1);
	}
}

void	ExitApplication( return_code )
int	return_code;
{
	LinkFree( LinkId, lockFileFlag );
	exit( return_code );
}

/*---------------------------------------------------------------------*/

void	LadeProgramm()
{
	switch ( LoadProgram( LinkId, 50, occamName ) ) {
		case -1:
			fprintf( stderr, "Fehler: Occam-Programm nicht gefunden; '%s'\n", occamName );
			ExitApplication( 1 );
			break;
		case -2:
			Error( "Fehler: Kann Occam-Programm nicht laden\n" );
			ExitApplication( 1 );
			break;
		case 0:
			break;
	}

}

/*------------------------------------------------------------------------------------*/
int	mode;				/* Kodierung der Instanzart. */
float	optimumKosten;			/* Die global besten Kosten der Instanz */
float	*xkoord = NULL;
float	*ykoord = NULL;

void	EndeLeseInstanz()
{
	free( xkoord ); free( ykoord );
}

int	LeseInstanz()
/*
	return: 0 => kein Felher aufgetreten
*/
{
	FILE	*f;
	int	i;

	if ( (f = fopen( instanzName, "r" )) == NULL ) {
		Error( "Fehler: Datei der Instanz nicht gefunden '%s'.\n", instanzName );
		return -1;
	}
	fscanf( f, "%d", &n );
	fscanf( f, "%f", &optimumKosten );
	fscanf( f, "%d", &mode );
	switch ( mode ) {
		case 1:
			break;
		case 10:
			break;
		default:
			Error( "Fehler: Falsche Instanzart\n" );
			return -1;
			break;
	}
	xkoord = (float*)malloc( n * sizeof( float ) );
	ykoord = (float*)malloc( n * sizeof( float ) );
	for ( i=0; i<n; i++ ) {
		fscanf( f, "%f", &xkoord[i] );
		fscanf( f, "%f", &ykoord[i] );
	}
	fclose( f );
	return 0;
}

int	SendeInstanz()
{
	int	i;

if ( debugflag ) printf("Sende die Instanz ...");

	if ( SendProt( LinkId, 50, TOWNPROTOCOL ) ) {
		Error( "Fehler: OCCAM Programm empfaengt die Instanz nicht\n");
		return -1;
	}
	if ( SendInt( LinkId, 50, &n ) != 0 ) {
		return -1;
	}
	if ( SendInt( LinkId, 50, &mode ) != 0 ) {
		return -1;
	}
	if ( SendReal( LinkId, 50, &optimumKosten ) != 0 ) {
		return -1;
	}
	for (i=0; i < n; i++ )
		if ( SendReal( LinkId, 50, &xkoord[i] ) != 0 ) {
			return -1;
		}
	for (i=0; i < n; i++ )
		if ( SendReal( LinkId, 50, &ykoord[i] ) != 0 ) {
			return -1;
		}
if ( debugflag ) printf("OK!!\n");
	return 0;
}

void	InitialisiereTransputer()
{
	LadeProgramm();					/* das Occamprogramm wird ins Netz geladen */
	if ( OccamVersion() < 0 ) ExitApplication( 1 );	/* teste, ob richtige Version des Occamprogramms */
	if ( LeseInstanz() < 0 ) ExitApplication( 1 );	/* Lese Instanzdatei */
	if ( SendeInstanz() < 0 ) ExitApplication( 1 );	/* Schicke Instanz ins Netz */
	EndeLeseInstanz();				/* Loesche Instanz aus C-Programm */
}

/*---------------------------------------------------------------------*/

int	SendeParameter()
{
if ( debugflag ) printf("Sende Parameter\n");
	if ( SendProt( LinkId, 50, PARAMPROTOCOL ) ) {
		Error( "Fehler: Parameter werden vom OCCAM Programm nicht empfangen\n" );
		return -1;
	}
	if ( SendInt( LinkId, 50, &anzRandomKonf ) ) {
		return -1;
	}
if ( debugflag ) printf("OK!!\n");
}

/*---------------------------------------------------------------------*/

int	EmpfangeStatistik()
{
	int	schritte;
	int	akzepts;
	int	non_akzepts;
	float	zeit;

	ReceiveInt( LinkId, 50, &schritte );
	Neuer_Wert( &g_schritte, (double)schritte );

	ReceiveInt( LinkId, 50, &akzepts );
	Neuer_Wert( &g_akzepts, (double)akzepts );

	ReceiveInt( LinkId, 50, &non_akzepts );
	Neuer_Wert( &g_nonakzepts, (double)non_akzepts );

	ReceiveReal( LinkId, 50, &zeit );
	Neuer_Wert( &g_zeit, (double)zeit );

	printf( "\n" );
	printf( "Anzahl der Schritte                  : %d\n", schritte );
	printf( "Anzahl der Akzeptierungen            : %d\n", akzepts );
	printf( "Anzahl der verworfenen Akzeptierungen: %d\n", non_akzepts );
	printf( "Zeit: %.4f\n", zeit );
}

/*---------------------------------------------------------------------*/

int	EmpfangeAlive()
{
	printf( "." );
	fflush( stdout );
	alive = 1;
}

/*---------------------------------------------------------------------*/

int	EmpfangeErgebnis()
{
	float	tour_Laenge;
	int	n;

	ReceiveReal( LinkId, 50, &tour_Laenge );
	Neuer_Wert( &g_wert, (double)tour_Laenge );
	/* vielleicht auch noch die ganze Tour */
	printf( "Tour hat die Laenge: %.5f\n", tour_Laenge );
	printf( "\n" );
}

/*---------------------------------------------------------------------*/

void	ReadProtocol( LinkId, run  )
int	LinkId;
int	*run;
{
	char	prot;

	ReceiveProt( LinkId, 50, &prot );
if ( debugflag ) printf("Protokoll: %d \n", prot );
	switch ( prot ) {
		case STATISTIC:
			EmpfangeStatistik();
			break;
		case ALIVE:
			EmpfangeAlive();
			break;
		case RESULT:
			EmpfangeErgebnis();
			break;
		case NETSTARTED:
			break;
		case NETTERMINATED:
			*run = FALSE;
			break;
		default:
			Error( "error: protocol (%d)\n", prot );
			*run = FALSE;
			break;
	}
}

/*---------------------------------------------------------------------*/

void	polling( run )
int	*run;
{
	if ( TestRead( LinkId ) ==  1 )
		ReadProtocol( LinkId, run );
	else
		*run = TRUE;
}

/*---------------------------------------------------------------------*/

void	AliveHandler( sig )
int	sig;
{
	if ( alive == 0 ) {
	/* Waehrend dieser Zeit kam keine Lebensmeldung vom Occamprogramm */
		ExitApplication( 2 );
	} else
		alive = 0;

	signal( sig, AliveHandler );
	alarm( 60 * ALARM );		/* in ALARM min Interrupt */
}

/*---------------------------------------------------------------------*/

void	main( argc, argv )
int	argc;
char	**argv;
{
	int	lauf;

	if ( strcmp( argv[argc-1], "-d" ) == 0 ) {
		argc--;
		debugflag = TRUE;
	}
	if ( strcmp( argv[argc-1], "-l" ) == 0 ) {
		argc--;
		lockFileFlag = FALSE;
	}
	if ( argc != 6 )
		usage( argv[0] );
	else {
		LinkId = atoi( argv[1] );
		strcpy( occamName, argv[2] );
		strcpy( instanzName, argv[3] );
		anzRandomKonf = atoi( argv[4] );
		anz_Laeufe = atoi( argv[5] );
	}

	SetzeLink();

	Init_Struktur( &g_wert );
	Init_Struktur( &g_schritte );
	Init_Struktur( &g_akzepts );
	Init_Struktur( &g_nonakzepts );
	Init_Struktur( &g_zeit );

		signal( SIGALRM, AliveHandler );
	for ( lauf=0; lauf < anz_Laeufe; lauf++ ) {
		int	running;

		InitialisiereTransputer();

		SendeParameter();

		alarm( 60 * ALARM );		/* in ALARM min Interrupt */

		running = TRUE;
		while ( running )
			polling( &running );
	}

	printf( "\n" );
	printf( "Prozessoren: %d; Laeufe: %d  ONE-CHAIN\n",
		anz_Proz, anz_Laeufe );
	printf( "Mittelwert++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" );
	printf( "Wert:       %.4lf %.2lf%%	(%.4lf)\n", Mittelwert( &g_wert ),
		((Mittelwert( &g_wert )-optimumKosten)/optimumKosten)*100.0, Sigma( &g_wert ) );
	printf( "Zeit:       %.4lf		(%.4lf)\n", Mittelwert( &g_zeit ), Sigma( &g_zeit ) );
	printf( "Schritte:   %.4lf	(%.4lf)\n", Mittelwert( &g_schritte ), Sigma( &g_schritte ) );
	printf( "akzepts:    %.4lf		(%.4lf)\n", Mittelwert( &g_akzepts ), Sigma( &g_akzepts ) );
	printf( "nonakzepts: %.4lf		(%.4lf)\n", Mittelwert( &g_nonakzepts ), Sigma( &g_nonakzepts ) );
	printf( "\n" );

	ExitApplication( 0 );
}

