/*************************************************************************

		tsp.c, v 1.00  90/10/28  16:57:43   author CH

**************************************************************************/

#include <stdio.h> 
#include "tsp.h"

extern	double	atof();
extern	void	exit();

/*-------------------------------------------*/
/* Variablen */

int	LinkId;

int	n;
float	Temp;
float	Alpha;
int	Random;
int	Laenge;
int	Tiefe;

char	graphName[200] = "instanzen/tsp442.instanz";
char	occamName[200] = "occam_src/xx";

int	debugflag = FALSE;

/*------------------------------------------------------------------------------------------------------------*/

void	Error( str )
char	*str;
{
	fprintf( stderr, "%s\n", str );
}

void usage( name )
char	*name;
{
	fprintf(stderr,
	"usage: %s	linkname [-fn {font}] [-update {seconds}] [-scale {integer}] [-rv]\n", name );
	fprintf(stderr,
	"		[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]] [-display [{host}]:[{vs}]]\n" );
	fprintf(stderr,
	"		[-fg {color}] [-bg {color}] [-bd {color}] [-bw {pixels}]\n");
	exit(1);
}

void	SetzeLink()
{
	int	err;
	char	ErrMsg[100];

	err = LinkAlloc( LinkId, ErrMsg );
	if (err != LinkId ) {
		fprintf( stderr, "%s\n", ErrMsg );
		exit(1);
	}
}

void	ExitApplication( return_code )
int	return_code;
{
	LinkFree( LinkId );
	exit( return_code );
}

void	LadeProgramm()
{
	switch ( LoadProgram( LinkId, 50, occamName ) ) {
		case -1:
			fprintf( stderr, "Fehler: Occam-Programm nicht gefunden; '%s'\n", occamName );
			ExitApplication( 1 );
			break;
		case -2:
			Error( "Fehler: Kann Occam-Programm nicht laden\n" );
			ExitApplication( 1 );
			break;
		case 0:
			break;
	}

}

/*------------------------------------------------------------------------------------*/
int	mode;				/* Kodierung der Instanzart. */
float	optimumKosten;			/* Die global besten Kosten der Instanz */
float	*xkoord = NULL;
float	*ykoord = NULL;

void	EndeLeseInstanz()
{
	free( xkoord ); free( ykoord );
}

int	LeseInstanz( dateiName )
char		*dateiName;
/*
	return: 0 => kein Felher aufgetreten
*/
{
	FILE	*f;
	int	i;

	if ( (f = fopen( dateiName, "r" )) == NULL ) {
		Error( "Fehler: Datei der Instanz nicht gefunden.\n" );
		return -1;
	}
	fscanf( f, "%d", &n );
	fscanf( f, "%f", &optimumKosten );
	fscanf( f, "%d", &mode );
	switch ( mode ) {
		case 1:
			break;
		case 10:
			break;
		default:
			Error( "Fehler: Falsche Instanzart\n" );
			return -1;
			break;
	}
	xkoord = (float*)malloc( n * sizeof( float ) );
	ykoord = (float*)malloc( n * sizeof( float ) );
	for ( i=0; i<n; i++ ) {
		fscanf( f, "%f", &xkoord[i] );
		fscanf( f, "%f", &ykoord[i] );
	}
	fclose( f );
	return 0;
}

int	SendeInstanz()
{
	int	i;

	if ( SendProt( LinkId, 50, PARAMPROTOCOL ) ) {
		Error( "Fehler: Parameter werden vom OCCAM Programm nicht empfangen\n" );
		return -1;
	}
	if ( SendReal( LinkId, 50, &Temp ) ) {
		return -1;
	}
	if ( SendInt( LinkId, 50, &Laenge ) ) {
		return -1;
	}
	if ( SendInt( LinkId, 50, &Tiefe ) ) {
		return -1;
	}
	if ( SendReal( LinkId, 50, &Alpha ) ) {
		return -1;
	}
	if ( SendInt( LinkId, 50, &Random ) ) {
		return -1;
	}
if ( debugflag ) printf("OK!!\n");
if ( debugflag ) printf("Sende die Instanz ...");

	if ( SendProt( LinkId, 50, TOWNPROTOCOL ) ) {
		Error( "Fehler: OCCAM Programm empfaengt die Instanz nicht\n");
		return -1;
	}
	if ( SendInt( LinkId, 50, &n ) != 0 ) {
		return -1;
	}
	if ( SendInt( LinkId, 50, &mode ) != 0 ) {
		return -1;
	}
	if ( SendReal( LinkId, 50, &optimumKosten ) != 0 ) {
		return -1;
	}
	for (i=0; i < n; i++ )
		if ( SendReal( LinkId, 50, &xkoord[i] ) != 0 ) {
			return -1;
		}
	for (i=0; i < n; i++ )
		if ( SendReal( LinkId, 50, &ykoord[i] ) != 0 ) {
			return -1;
		}
if ( debugflag ) printf("OK!!\n");
	return 0;
}

/*---------------------------------------------------------------------*/

int	EmpfangeTour()
{
	int	tour[800];
	int	n, i;

if ( debugflag ) printf( "receive tour " );
	ReceiveInt( LinkId, 50, &n );

if ( debugflag ) printf("n: %d .... ", n );
	for (i=0; i<n; i++)
		ReceiveInt( LinkId, 50, &tour[i] );

if ( debugflag ) printf( " Ok!\n" );
	return 0;
}

/*---------------------------------------------------------------------*/

int	EmpfangeTourlen()
{
	float	Tour_Laenge;

	ReceiveReal( LinkId, 50, &Tour_Laenge );
if ( debugflag ) printf("Tour hat die Laenge: %f\n", Tour_Laenge );
}

/*---------------------------------------------------------------------*/

int	EmpfangeTemperatur()
{
	float	Temperatur;

	ReceiveReal( LinkId, 50, &Temperatur );
if ( debugflag ) printf("Temperatur: %f\n", Temperatur );
}

/*---------------------------------------------------------------------*/

int	EmpfangeAkzeptierungen()
{
	int	accept;

	ReceiveInt( LinkId, 50, &accept );
if ( debugflag ) printf("Akzeptierungen: %d\n", accept );
}

/*---------------------------------------------------------------------*/

void	EmpfangeVersion()
{
	int	version;

	ReceiveInt( LinkId, 50, &version );
	if ( version != 1 )
		Error( "Fehler: OCCAM Programm sendet FALSCHE Versionsnummer\n" );
if ( debugflag ) printf("Versionsnummer %d:\n", version );
}

/*---------------------------------------------------------------------*/

void	ReadProtocol( LinkId )
int	LinkId;
{
	char	prot;

	ReceiveProt( LinkId, 50, &prot );
if ( debugflag ) printf("Protokoll: %d \n", prot );
	switch ( prot ) {
		case TOURPROTOCOL:
			EmpfangeTour();
			break;
		case STATISTICTOURLEN:
			EmpfangeTourlen();
			break;
		case STATISTICTEMP:
			EmpfangeTemperatur();
			break;
		case STATISTICACCEPT:
			EmpfangeAkzeptierungen();
			break;
		case NETSTARTED:
			break;
		case NETTERMINATED:
			break;
		case VERSION:
			EmpfangeVersion();
			break;
		default:
			Error( "error: protocol (%d)\n", prot );
	}
}

/*---------------------------------------------------------------------*/

int	OccamVersion()
{
	int	version = 1;


	if ( SendInt( LinkId, 50, &version ) != 0 ) {
		Error( "Fehler: OCCAM Programm empfaengt keine Versionsnummer\n");
		return -1;
	}
	if ( ReceiveInt( LinkId, 50, &version ) != 0 ) {
		Error( "Fehler: OCCAM Programm sendet keine Versionsnummer\n");
		return -1;
	}
	if ( version != 1 ) {
		Error( "Fehler: OCCAM Programm sendet FALSCHE Versionsnummer\n" );
		return -1;
	} else
		return 0;
}

/*---------------------------------------------------------------------*/

void	polling()
{
	if ( TestRead( LinkId ) ==  1 )
		ReadProtocol( LinkId );
}

/*---------------------------------------------------------------------*/

void main(argc, argv)
    int argc;
    char **argv;
{
	printf( "Linknummer: " );
	scanf( "%d", &LinkId );

	SetzeLink();

	LadeProgramm();

	while (TRUE)
		polling();

	ExitApplication( 0 );
}

