/***********************************************************************

		Diagram.c, v 1.1  90/2/12  13:21:43   author CH

***********************************************************************

/***********************************************************
Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#include <X11/IntrinsicP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Xmu.h>
#include <X11/cursorfont.h>
#include <stdio.h>

#include "DiagramP.h"
#include "DiagramI.h"

/* Private Data */

#define	offset(field)	XtOffset( DiagramWidget,diagram.field )
#define	goffset(field)	XtOffset( Widget,core.field )

static	char	defaultTranslations[] =
	"<EnterWindow>:		enter_window()		\n\
	<LeaveWindow>:		leave_window()		";

static	XtActionsRec	actionsList[] =
{
	{"enter_window",	Enter_Window},
	{"leave_window",	Leave_Window},
};

static	Dimension	defDimWidth = 200;
static	Dimension	defDimHeight = 100;

static	XtResource	resources[] = {
	{XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
		goffset( width ), XtRDimension, (caddr_t)&defDimWidth},
	{XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
		goffset( height ), XtRDimension, (caddr_t)&defDimHeight},
	{XtNbackground, XtCBackground, XtRPixel, sizeof( Pixel ),
		goffset( background_pixel ), XtRString, "White"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof( Pixel ),
       		offset( fgpixel ), XtRString, "Black"},
	{XtNhighlight, XtCForeground, XtRPixel, sizeof( Pixel ),
		offset( hipixel ), XtRString, "BLACK" },
	{XtNfont, XtCFont, XtRFontStruct, sizeof( XFontStruct * ),
       		offset( font ), XtRString, XtDefaultFont},
	{XtNlabel, XtCLabel, XtRString, sizeof( char * ),
		offset( text ), XtRString, NULL },
};

#undef	offset
#undef	goffset

DiagramClassRec diagramClassRec = {
    { /* core fields */
    /* superclass		*/	&widgetClassRec,
    /* class_name		*/	"Diagram",
    /* size			*/	sizeof(DiagramRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actionsList,
    /* num_actions		*/	XtNumber( actionsList ),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULL,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	NULL,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	defaultTranslations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
    }
};

WidgetClass diagramWidgetClass = (WidgetClass) &diagramClassRec;

/****************************************************************
 *
 * Private Procedures
 *
 ****************************************************************/

static void ClassInitialize()
{
    XtAddConverter( XtRFunction, XtRCallback, XmuCvtFunctionToCallback,
		    NULL, 0 );
}

/* ARGSUSED */
static	void	Initialize (greq, gnew)
Widget	greq, gnew;
{
	DiagramWidget	w = (DiagramWidget)gnew;
	XtGCMask	valuemask;
	XGCValues	myXGCV;

	w->diagram.cursor = XCreateFontCursor( XtDisplay (w), XC_gumby );
	if (w->diagram.cursor == None )
		fprintf ( stderr, "unable to grab cursor\n" );

	valuemask = GCForeground | GCBackground;
	myXGCV.foreground = w->diagram.fgpixel;
	if (w->diagram.font) {
		myXGCV.font = w->diagram.font->fid;
		valuemask |= GCFont;
	}
	myXGCV.background = w->core.background_pixel;
	w->diagram.fgGC = XtGetGC( gnew, valuemask, &myXGCV );
	myXGCV.foreground = w->core.background_pixel;
	w->diagram.clGC = XtGetGC( gnew, valuemask, &myXGCV );
	myXGCV.foreground = w->diagram.hipixel;
	w->diagram.hiGC = XtGetGC( gnew, valuemask, &myXGCV );

	/*
	* Note that the second argument is a GCid -- QueryFont accepts a GCid and
	* returns the curently contained font.
	*/
	if (w->diagram.font == NULL)
		w->diagram.font = XQueryFont (XtDisplay (w),
      			XGContextFromGC (DefaultGCOfScreen (XtScreen (w))));

	/*
	* initialize variable
	*/
	w->diagram.n = 0;			/* Kein Zahlenwert in der Tabelle */	
	strcpy( w->diagram.valstr, "" );
	w->diagram.x_val = XTEXT + 20 +
			XTextWidth( w->diagram.font, w->diagram.text, strlen(w->diagram.text) );
	w->diagram.y_val = YTEXT + w->diagram.font->ascent;
	w->diagram.y_offset = w->diagram.y_val + 2;

	w->diagram.max_value = 0.0;
	w->diagram.scale = 1;
	w->diagram.x_n = 0;
}
 
static	void	Realize (gw, valueMask, attrs)
Widget 			gw;
XtValueMask		*valueMask;
XSetWindowAttributes	*attrs;
{
	XtCreateWindow( gw, (unsigned)InputOutput, (Visual *)CopyFromParent,
		*valueMask, attrs );
}

static	void	Destroy (gw)
Widget	gw;
{
     DiagramWidget w = (DiagramWidget)gw;

     XFreeCursor( XtDisplay( w ), w->diagram.cursor );
     XtDestroyGC (w->diagram.fgGC);
     XtDestroyGC (w->diagram.clGC);
}

static	void	Resize( gw )
Widget	gw;
{
	DiagramWidget	w = (DiagramWidget)gw;

	repaint_window( w );
}

/* ARGSUSED */
static void Redisplay(gw, event, region)
     Widget gw;
     XEvent *event;
     Region region;
{
	DiagramWidget	w = (DiagramWidget)gw;

	repaint_window( w );
}

/* ARGSUSED */
static Boolean SetValues (current, request, new)
    Widget current, request, new;
{
    return( FALSE );
}

static	void	Enter_Window( w, event, params, num_params )
Widget		w;
XEvent		*event;
String		*params;		/* unused */
Cardinal	*num_params;		/* unused */
{
	DiagramWidget	diagram = (DiagramWidget)w;

	XDefineCursor( XtDisplay( diagram ), XtWindow( diagram ), diagram->diagram.cursor );
}

static	void	Leave_Window( w, event, params, num_params )
Widget		w;
XEvent		*event;
String		*params;		/* unused */
Cardinal	*num_params;		/* unused */
{
	DiagramWidget	diagram = (DiagramWidget)w;

	XUndefineCursor( XtDisplay( diagram ), XtWindow( diagram ) );
}

/*--------------------------------------------------------------------------------------------------*/

static	void	repaint_window( w )
DiagramWidget	w;
{
	Display	*dpy = XtDisplay( w );
	Window	win = XtWindow( w );

	/* Falls Window offen */
	if ( XtIsRealized( (Widget)w ) ) {
		int	i;
		int	y_line;

		/* Loesche gesamtes Window */
		XClearWindow( XtDisplay( w ), XtWindow( w ) );
		/* Schreibe Windownamen */
		if ( w->diagram.text )
			XDrawString( dpy, win, w->diagram.hiGC, XTEXT, YTEXT + w->diagram.font->ascent,
				w->diagram.text, strlen( w->diagram.text ) );
		/* Falls ein Wert in der Tabelle */
		if ( w->diagram.n > 0 ) {
			/* Schreibe Zahlenwert */
			sprintf( w->diagram.valstr, "%f", w->diagram.value[w->diagram.n - 1] );
			XDrawString( dpy, win, w->diagram.hiGC, w->diagram.x_val, w->diagram.y_val,
				 w->diagram.valstr, strlen( w->diagram.valstr ) );
			/* Zeichne Diagrammlinien */
			w->diagram.x_n = 0;
			for ( i=0; i<w->diagram.n; i++ ) {
				y_line = (int)(w->core.height - 
					((w->core.height-w->diagram.y_offset)*w->diagram.value[i]) / w->diagram.scale );
				XDrawLine( dpy, win, w->diagram.fgGC,
					w->diagram.x_n, w->core.height,
					w->diagram.x_n, y_line );
				w->diagram.x_n++;
			}
		}
	}
}

static	void	draw_it( w )
DiagramWidget	w;
{
	float	value;

	value = w->diagram.value[w->diagram.n - 1];
	/* Behalte max_value korrekt */
	if ( value > w->diagram.max_value ) {
		w->diagram.max_value = value;
		w->diagram.scale = ( (int)w->diagram.max_value ) + 1 ;
		repaint_window( w );
		return;
	}

	if ( XtIsRealized( (Widget)w ) ) {
		Display	*dpy = XtDisplay( w );
		Window	win = XtWindow( w );

		/* Loesche alten Zahlenwert */
		if ( w->diagram.n > 1 ) {
			XDrawString( dpy, win, w->diagram.clGC, w->diagram.x_val, w->diagram.y_val,
				w->diagram.valstr, strlen( w->diagram.valstr ) );
		}
		/* Schreibe neuen Zahlenwert */
		sprintf( w->diagram.valstr, "%f", value );
		XDrawString( dpy, win, w->diagram.hiGC, w->diagram.x_val, w->diagram.y_val,
			w->diagram.valstr, strlen( w->diagram.valstr ) );
		/* Zeichne Diagramlinie */
		XDrawLine( dpy, win, w->diagram.fgGC,
			w->diagram.x_n, w->core.height,
			w->diagram.x_n,
			(int)(w->core.height - ((w->core.height-w->diagram.y_offset)*value) / w->diagram.scale) );
/*
		XFlush( dpy );
*/
		w->diagram.x_n++;
	}
}

/***************************************************************************************
 The following routines provide procedural interfaces to Diagram Window
statesetting and getting.
****************************************************************************************/

int	XtSetDiagramValue( gw, f )
Widget	gw;
double	f;
{
	DiagramWidget	w = (DiagramWidget)gw;

	w->diagram.value[w->diagram.n] = (float)f;
	(w->diagram.n)++;
	draw_it( w );
}

