/***********************************************************
Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#include "TimerP.h"

extern long time();
static void timer_tic();

/* Provate Definitions */

static	char*	diff_time( long t );

/* Initialization of defaults */

#define	offset(field)	XtOffset(TimerWidget,timer.field)
#define	goffset(field)	XtOffset(Widget,core.field)

static XtResource resources[] = {
	{XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
		goffset(width), XtRString, "0"},
	{XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
		goffset(height), XtRString, "0"},
	{XtNbackground, XtCBackground, XtRPixel, sizeof(Pixel),
		goffset(background_pixel), XtRString, "white"},
	{XtNupdate, XtCInterval, XtRInt, sizeof(int), 
       		offset(update), XtRString, "5" },
	{XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
       		offset(fgpixel), XtRString, "black"},
	{XtNhighlight, XtCForeground, XtRPixel, sizeof(Pixel),
       		offset(Hipixel), XtRString, "black"},
	{XtNpadding, XtCMargin, XtRInt, sizeof(int),
       		offset(padding), XtRString, "8"},
	{XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
       		offset(font), XtRString, "fixed"},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
		offset (reverse_video), XtRString, "FALSE"},
	{XtNbackingStore, XtCBackingStore, XtRBackingStore, sizeof( int ),
		offset( backing_store ), XtRString, "default" },
};

#undef offset
#undef goffset

static void ClassInitialize();
static void Initialize(), Realize(), Destroy(), Resize(), Redisplay();
static Boolean SetValues();

TimerClassRec timerClassRec = {
    { /* core fields */
    /* superclass		*/	&widgetClassRec,
    /* class_name		*/	"Timer",
    /* widget_size		*/	sizeof(TimerRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	resources,
    /* resource_count		*/	XtNumber(resources),
    /* xrm_class		*/	NULL,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry           */	XtInheritQueryGeometry,
    /* display_accelerator      */	XtInheritDisplayAccelerator,
    /* extension                */	NULL
    }
};

WidgetClass timerWidgetClass = (WidgetClass) &timerClassRec;

/****************************************************************
 *
 * Private Procedures
 *
 ****************************************************************/

static void ClassInitialize()
{
    XtAddConverter( XtRString, XtRBackingStore, XmuCvtStringToBackingStore,
		    NULL, 0 );
}

/* ARGSUSED */
static	void	Initialize( request, new )
Widget	request, new;
{
	TimerWidget	w = (TimerWidget)new;
	XtGCMask	valuemask;
	XGCValues	myXGCV;
	char	*str;
	int		min_height, min_width;

	valuemask = GCForeground | GCBackground | GCFont | GCLineWidth;
	if (w->timer.font != NULL)
		myXGCV.font = w->timer.font->fid;
	else
		valuemask &= ~GCFont;	/* use server default font */

	str = diff_time( 0 );
	if (w->timer.font == NULL)
		w->timer.font = XQueryFont( XtDisplay(w),
					XGContextFromGC(
					DefaultGCOfScreen(XtScreen(w))) );
	min_width = XTextWidth(w->timer.font, str, strlen(str)) +
			2 * w->timer.padding;
	min_height = w->timer.font->ascent +
			w->timer.font->descent + 2 * w->timer.padding;
    if (w->core.width == 0)
	w->core.width = min_width;
    if (w->core.height == 0)
	w->core.height = min_height;

    /*
     * set the colors if reverse video; this is somewhat tricky since there
     * are 5 colors:
     *
     *     background - paper		white
     *     foreground - text, ticks	black
     *     border - border		black (foreground)
     *     highlight - edge of hands	black (foreground)
     *
     * This doesn't completely work since the parent has already made up a 
     * border.  Sigh.
     */
    if (w->timer.reverse_video) {
	Pixel fg = w->timer.fgpixel;
	Pixel bg = w->core.background_pixel;

	if (w->timer.Hipixel == fg) w->timer.Hipixel = bg;
	if (w->core.border_pixel == fg) w->core.border_pixel = bg;
	w->timer.fgpixel = bg;
	w->core.background_pixel = fg;
    }

    myXGCV.foreground = w->timer.fgpixel;
    myXGCV.background = w->core.background_pixel;
    if (w->timer.font != NULL)
        myXGCV.font = w->timer.font->fid;
    else
        valuemask &= ~GCFont;	/* use server default font */
    myXGCV.line_width = 0;
    w->timer.myGC = XtGetGC((Widget)w, valuemask, &myXGCV);

    valuemask = GCForeground | GCLineWidth ;
    myXGCV.foreground = w->core.background_pixel;
    w->timer.EraseGC = XtGetGC((Widget)w, valuemask, &myXGCV);

    myXGCV.foreground = w->timer.Hipixel;
    w->timer.HighGC = XtGetGC((Widget)w, valuemask, &myXGCV);

    if (w->timer.update <= 0)
	w->timer.update = 60;	/* make invalid update's use a default */
    w->timer.interval_id = 0;
    w->timer.start_time = 0;
    w->timer.running = FALSE;
}

static void Realize (gw, valueMask, attrs)
     Widget gw;
     XtValueMask *valueMask;
     XSetWindowAttributes *attrs;
{
     TimerWidget	w = (TimerWidget) gw;
#ifdef notdef
     *valueMask |= CWBitGravity;
     attrs->bit_gravity = ForgetGravity;
#endif
     switch (w->timer.backing_store) {
     case Always:
     case NotUseful:
     case WhenMapped:
     	*valueMask |=CWBackingStore;
	attrs->backing_store = w->timer.backing_store;
	break;
     }
     XtCreateWindow( gw, InputOutput, (Visual *)CopyFromParent,
		     *valueMask, attrs);
     Resize( gw );
}

static void Destroy (gw)
     Widget gw;
{
     TimerWidget w = (TimerWidget) gw;
     if (w->timer.interval_id) XtRemoveTimeOut (w->timer.interval_id);
     XtDestroyGC (w->timer.myGC);
     XtDestroyGC (w->timer.HighGC);
     XtDestroyGC (w->timer.EraseGC);
}

static void Resize (gw) 
    Widget gw;
{
    TimerWidget w = (TimerWidget) gw;
    /* don't do this computation if window hasn't been realized yet. */
}

char*	diff_time( t )
long	t;
{
	struct tm	*stime;
	static	char	str[20];

	stime = gmtime( &t );
	sprintf( str, "%02d:%02d:%02d", stime->tm_hour, stime->tm_min, stime->tm_sec );
	return str;
}

void	write_time( w )
TimerWidget	w;
{
	Display	*dpy = XtDisplay( w );
	Window	win = XtWindow( w );
	long	time_diff;
	char	*time_ptr;

	if ( w->timer.running ) {
		long	time_value;

		(void) time(&time_value);
		time_diff = time_value - w->timer.start_time;
	}
	else
		time_diff = 0;
	time_ptr = diff_time( time_diff );
	XDrawImageString( dpy, win, w->timer.myGC,
			2+w->timer.padding, 2+w->timer.font->ascent+w->timer.padding,
			time_ptr, strlen( time_ptr ) );
}

/* ARGSUSED */
static void Redisplay (gw, event, region)
    Widget gw;
    XEvent *event;		/* unused */
    Region region;		/* unused */
{
	TimerWidget	w = (TimerWidget) gw;
	char	*time_ptr;
	Display	*dpy = XtDisplay( w );
	Window	win = XtWindow( w );

	write_time( w );
}

/* ARGSUSED */
static	void	timer_tic( client_data, id )
caddr_t		client_data;
XtIntervalId	*id;
{
        TimerWidget	w = (TimerWidget)client_data;    
/*
	long		time_value;
	long		time_diff;
	char		*time_ptr;
        register	Display *dpy = XtDisplay(w);
        register	Window win = XtWindow(w);
*/

	if ( !w->timer.running ) return;

	w->timer.interval_id =
		XtAddTimeOut( w->timer.update*1000, timer_tic, (caddr_t)w );

	write_time( w );
/*
	(void) time(&time_value);
	time_diff = time_value - w->timer.start_time;

	time_ptr = diff_time( time_diff );
	XDrawImageString (dpy, win, w->timer.myGC,
			2+w->timer.padding, 2+w->timer.font->ascent+w->timer.padding,
			time_ptr, strlen( time_ptr ) );
*/
}
	

/* ARGSUSED */
static Boolean SetValues (gcurrent, grequest, gnew)
    Widget gcurrent, grequest, gnew;
{
      TimerWidget current = (TimerWidget) gcurrent;
      TimerWidget new = (TimerWidget) gnew;
      Boolean redisplay = FALSE;
      XtGCMask valuemask;
      XGCValues	myXGCV;

      /* first check for changes to timer-specific resources.  We'll accept all
         the changes, but may need to do some computations first. */

      if ( new->timer.update != current->timer.update ) {
	  if ( current->timer.interval_id )
	      XtRemoveTimeOut( current->timer.interval_id );
	  if ( new->timer.running )
	      new->timer.interval_id = XtAddTimeOut( new->timer.update*1000,
						    timer_tic, (caddr_t)gnew );
      }

      if (new->timer.padding != current->timer.padding)
	   redisplay = TRUE;

       if (new->timer.font != current->timer.font)
	   redisplay = TRUE;

      if ((new->timer.fgpixel != current->timer.fgpixel)
          || (new->core.background_pixel != current->core.background_pixel)) {
          valuemask = GCForeground | GCBackground | GCFont | GCLineWidth;
	  myXGCV.foreground = new->timer.fgpixel;
	  myXGCV.background = new->core.background_pixel;
          myXGCV.font = new->timer.font->fid;
	  myXGCV.line_width = 0;
	  XtDestroyGC (current->timer.myGC);
	  new->timer.myGC = XtGetGC(gcurrent, valuemask, &myXGCV);
	  redisplay = TRUE;
          }

      if (new->timer.Hipixel != current->timer.Hipixel) {
          valuemask = GCForeground | GCLineWidth;
	  myXGCV.foreground = new->timer.fgpixel;
          myXGCV.font = new->timer.font->fid;
	  myXGCV.line_width = 0;
	  XtDestroyGC (current->timer.HighGC);
	  new->timer.HighGC = XtGetGC((Widget)gcurrent, valuemask, &myXGCV);
	  redisplay = TRUE;
          }

      if (new->core.background_pixel != current->core.background_pixel) {
          valuemask = GCForeground | GCLineWidth;
	  myXGCV.foreground = new->core.background_pixel;
	  myXGCV.line_width = 0;
	  XtDestroyGC (current->timer.EraseGC);
	  new->timer.EraseGC = XtGetGC((Widget)gcurrent, valuemask, &myXGCV);
	  redisplay = TRUE;
	  }
     
     return (redisplay);

}

/*******************************************************************************
*
*
*
*******************************************************************************/

void	XtSetStartTimer( w )
TimerWidget	w;
{
	w->timer.running = TRUE;
	(void)time( &(w->timer.start_time) );		/* set timer at time */
	w->timer.prev_time_string[0] = '\0';
	timer_tic((caddr_t)w, (XtIntervalId)0);
}

long	XtGetTimer( w )
TimerWidget	w;
{
	long	time_value;

	(void)time( &time_value );
	if ( w->timer.running )
		return  ( time_value - w->timer.start_time );
	else
		return 0;

}

long	XtSetStopTimer( w )
TimerWidget	w;
{
	long		time_value;

	(void)time( &time_value );
	if ( w->timer.running ) {
		if (w->timer.interval_id) XtRemoveTimeOut (w->timer.interval_id);
		w->timer.interval_id = 0;
		write_time( w );
		w->timer.running = FALSE;
		return	( time_value - w->timer.start_time );
	}
	else
		return 0;
}
