/***********************************************************************

		Tsp.c, v 1.10  90/2/28  10:30:43   author CH

***********************************************************************

/***********************************************************
Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#include <X11/IntrinsicP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Xmu.h>
#include <X11/cursorfont.h>
#include <stdio.h>

#include "TspP.h"
#include "TspI.h"

extern	void	exit();

/* Private Data */

#define offset(field) XtOffset(TspWidget,tsp.field)
#define goffset(field) XtOffset(Widget,core.field)

static	char	defaultTranslations[] =
	"<EnterWindow>:		enter_window()		\n\
	<LeaveWindow>:		leave_window()		";

static	XtActionsRec	actionsList[] =
{
	{"enter_window",	Enter_Window},
	{"leave_window",	Leave_Window},
};

static Dimension defDim = 300;

static XtResource resources[] = {
	{XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
		goffset(width), XtRDimension, (caddr_t)&defDim},
	{XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
		goffset(height), XtRDimension, (caddr_t)&defDim},
	{XtNbackground, XtCBackground, XtRPixel, sizeof(Pixel),
		goffset(background_pixel), XtRString, "White"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
		offset(fgpixel), XtRString, "Black"},
	{XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
       		offset(font), XtRString, XtDefaultFont},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof(Boolean),
		offset(reverseVideo), XtRString, "FALSE"},
};

#undef offset
#undef goffset

TspClassRec tspClassRec = {
    { /* core fields */
    /* superclass		*/	&widgetClassRec,
    /* class_name		*/	"Tsp",
    /* size			*/	sizeof(TspRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actionsList,
    /* num_actions		*/	XtNumber( actionsList ),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULL,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	NULL,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	defaultTranslations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
    }
};

WidgetClass tspWidgetClass = (WidgetClass) &tspClassRec;

/****************************************************************
 *
 * Private Procedures
 *
 ****************************************************************/

static void ClassInitialize()
{
    XtAddConverter( XtRFunction, XtRCallback, XmuCvtFunctionToCallback,
		    NULL, 0 );
}

/* ARGSUSED */
static	void	Initialize (greq, gnew)
Widget	greq, gnew;
{
	TspWidget	w = (TspWidget)gnew;
	Display		*dpy = XtDisplay( w );
	Window		win = XtWindow( w );
	int		screen = DefaultScreen( dpy );
	int		depth = DefaultDepth( dpy, screen );
	Colormap	cmap;
	XColor		exact;
	XColor		color;
	XtGCMask	valuemask;
	XGCValues	myXGCV;
	Visual		*visual;
	int		class;

	w->tsp.cursor = XCreateFontCursor( dpy, XC_gumby );
	if (w->tsp.cursor == None )
		fprintf ( stderr, "unable to grab cursor\n" );

	visual = DefaultVisual( dpy, screen );
	class = visual->class;
	if ( depth == 1) {
		w->tsp.pseudoColor = FALSE;
		valuemask = GCForeground | GCBackground;
		if ( w->tsp.reverseVideo ) {
/*
			XSetWindowBackground( dpy, win, w->tsp.fgpixel );
*/
			myXGCV.background = w->tsp.fgpixel;
			myXGCV.foreground = w->core.background_pixel;
		} else {	
			myXGCV.background = w->core.background_pixel;
			myXGCV.foreground = w->tsp.fgpixel;
		}
		w->tsp.townGC = XtGetGC(gnew, valuemask, &myXGCV);		/* town GC */
		w->tsp.new_tour_GC = XtGetGC(gnew, valuemask, &myXGCV);		/* newtour GC */
		XSetLineAttributes( dpy, w->tsp.new_tour_GC, 0, LineOnOffDash, CapButt, JoinMiter );
		w->tsp.old_tour_GC = XtGetGC(gnew, valuemask, &myXGCV);		/* old tour GC */
		XSetLineAttributes( dpy, w->tsp.old_tour_GC, 0, LineSolid, CapButt, JoinMiter );
		myXGCV.foreground = myXGCV.background;
		w->tsp.clGC = XtGetGC(gnew, valuemask, &myXGCV);		/* loesche GC */
/*
		XSetFunction( dpy, w->tsp.tourGC, GXxor );
*/
	} else if ( class == PseudoColor ) {
		w->tsp.pseudoColor = TRUE;
		valuemask = GCForeground | GCBackground;
		myXGCV.background = WHITE_COLOR;
		myXGCV.foreground = GREEN_COLOR;
		w->tsp.townGC = XtGetGC(gnew, valuemask, &myXGCV);		/* town GC */
		myXGCV.background = WHITE_COLOR;
		myXGCV.foreground = GREEN_COLOR;
		w->tsp.new_tour_GC = XtGetGC(gnew, valuemask, &myXGCV);		/* tour new GC */
		myXGCV.background = WHITE_COLOR;
		myXGCV.foreground = RED_COLOR;
		w->tsp.old_tour_GC = XtGetGC(gnew, valuemask, &myXGCV);		/* tour old GC */
		myXGCV.background = BLACK_COLOR;
		myXGCV.foreground = BLACK_COLOR;
		w->tsp.clGC = XtGetGC(gnew, valuemask, &myXGCV);		/* loesche GC */
		XSetLineAttributes( dpy, w->tsp.clGC, 1, LineSolid, CapButt, JoinMiter );
	} else {
		fprintf( stderr,"graphicboard not implemented\n" );
		exit(1);
	}

	if ( w->tsp.font ) {
		myXGCV.font = w->tsp.font->fid;
		valuemask |= GCFont;
	}
	/*
	* Note that the second argument is a GCid -- QueryFont accepts a GCid and
	* returns the curently contained font.
	*/
	if (w->tsp.font == NULL)
		w->tsp.font = XQueryFont( dpy, XGContextFromGC( DefaultGCOfScreen( XtScreen (w) ) ) );

	/*
	* initialize variable
	*/
	w->tsp.n = 0;				/* Object is empty */
	w->tsp.tour_stored = FALSE;		/* No tour stored */
	w->tsp.old_tour_stored = FALSE;		/* No old tour stored */
	w->tsp.tour = (Neighbor*)XtCalloc( MAX_N, sizeof( Neighbor ) );
	w->tsp.old_tour = (Neighbor*)XtCalloc( MAX_N, sizeof( Neighbor ) );
	w->tsp.tour_buf = (Neighbor*)XtCalloc( MAX_N, sizeof( Neighbor ) );
}
 
static	void	Realize (gw, valueMask, attrs)
Widget 			gw;
XtValueMask		*valueMask;
XSetWindowAttributes	*attrs;
{
	XtCreateWindow( gw, (unsigned)InputOutput, (Visual *)CopyFromParent,
		*valueMask, attrs );
}

static	void	Destroy (gw)
Widget	gw;
{
	TspWidget w = (TspWidget)gw;

	XFreeCursor( XtDisplay( w ), w->tsp.cursor );
	XtDestroyGC (w->tsp.townGC);
	XtDestroyGC (w->tsp.new_tour_GC);
	XtDestroyGC (w->tsp.old_tour_GC);
	XtFree( w->tsp.tour );
	XtFree( w->tsp.old_tour );
	XtFree( w->tsp.tour_buf );
}

static	void	Resize( gw )
Widget	gw;
{
	TspWidget	w = (TspWidget)gw;

	repaint_window( w );
}

/* ARGSUSED */
static	void	Redisplay(gw, event, region)
Widget	gw;
XEvent	*event;
Region	region;
{
	repaint_window( gw );
}

/* ARGSUSED */
static	Boolean	SetValues( current, request, new )
Widget	current, request, new;
{
	return( FALSE );
}

static	void	Enter_Window( w, event, params, num_params )
Widget		w;
XEvent		*event;
String		*params;		/* unused */
Cardinal	*num_params;		/* unused */
{
	TspWidget	tsp = (TspWidget)w;

	XDefineCursor( XtDisplay( tsp ), XtWindow( tsp ), tsp->tsp.cursor );
}

static	void	Leave_Window( w, event, params, num_params )
Widget		w;
XEvent		*event;
String		*params;		/* unused */
Cardinal	*num_params;		/* unused */
{
	TspWidget	tsp = (TspWidget)w;

	XUndefineCursor( XtDisplay( tsp ), XtWindow( tsp ) );
}

/*--------------------------------------------------------------------------------------------------*/

static	void	repaint_window( w )
TspWidget	w;
{
	if ( w->tsp.reverseVideo )
		XSetWindowBackground( XtDisplay( w ), XtWindow( w ), w->tsp.fgpixel );
	XClearWindow( XtDisplay( w ), XtWindow( w ) );
	XtCalculatePos( w );
	XtDrawTowns( w );
	XtDrawTour( w );
}

/***************************************************************************************
 The following routines provide procedural interfaces to TSP window
statesetting and getting.
****************************************************************************************/

int	XtSetTownPositions( w, n, x, y )
TspWidget	w;
int	n;			/* Number of towns */
float	*x;		/* X-Coordinate of the town */
float	*y;			/* Y-Coordinate of the town */
{
	int	i;

	w->tsp.n = n;
	for (i=0; i < n; w->tsp.xcoord[i++] = *(x++) );
	for (i=0; i < n; w->tsp.ycoord[i++] = *(y++) );

	w->tsp.xmin = w->tsp.xcoord[0];
	w->tsp.xmax = w->tsp.xcoord[0];
	w->tsp.ymin = w->tsp.ycoord[0];
	w->tsp.ymax = w->tsp.ycoord[0];
	for (i=0; i < w->tsp.n; i++) {
		if ( w->tsp.xmax < w->tsp.xcoord[i] ) w->tsp.xmax = w->tsp.xcoord[i];
		if ( w->tsp.xmin > w->tsp.xcoord[i] ) w->tsp.xmin = w->tsp.xcoord[i];
		if ( w->tsp.ymax < w->tsp.ycoord[i] ) w->tsp.ymax = w->tsp.ycoord[i];
		if ( w->tsp.ymin > w->tsp.ycoord[i] ) w->tsp.ymin = w->tsp.ycoord[i];
	}

	w->tsp.tour_stored = FALSE;
	w->tsp.old_tour_stored = FALSE;

	return 0;
}

int	XtSetTour( w , t )
TspWidget	w;
int		*t;		/* Description of the tour */
{
	Neighbor	*p;
	int	i;
	int	n = w->tsp.n;

	if ( w->tsp.tour_stored == FALSE ) {
	/* The first tour is stored */
		w->tsp.tour_stored = TRUE;
		w->tsp.old_tour_stored = FALSE;
	} else {
	/* store the new tour. Old tour must be saved. */
		w->tsp.old_tour_stored = TRUE;
		p = w->tsp.old_tour;
		w->tsp.old_tour = w->tsp.tour;
		w->tsp.tour = p;
	}

	for ( i=0; i < n; i++ ) {
		w->tsp.tour[t[i]].t1 = t[(i+n-1) % n];
		w->tsp.tour[t[i]].t2 = t[(i+1) % n];
	}


	return 0;
}

#define	RAND	4

void	XtCalculatePos( w )
TspWidget	w;
{
	float	f1, f2, f;
	int	width = w->core.width - (2 * RAND);
	int	height = w->core.height - (2 * RAND);
	int	i;

	f1 = ( w->tsp.xmax -  w->tsp.xmin) / (float) width;
	f2 = ( w->tsp.ymax -  w->tsp.ymin) / (float) height;
	if (f1 < f2)
		f = f2;
	else
		f = f1;

	for ( i=0; i < w->tsp.n; i++ ) {
		w->tsp.xpos[i] = RAND + (int) (width - (w->tsp.xcoord[i] - w->tsp.xmin) / f);
		w->tsp.ypos[i] = RAND + (int) ((w->tsp.ycoord[i] - w->tsp.ymin) / f);
	}
}

void	XtDrawTowns( w )
TspWidget	w;
{
	int	i;

	if ( w->tsp.pseudoColor == FALSE )
		for (i=0; i < w->tsp.n; i++ ) {
			XFillRectangle( XtDisplay( w ), XtWindow( w ), w->tsp.townGC,
				w->tsp.xpos[i]-1, w->tsp.ypos[i]-1, 3, 3 );
	} else {
		for (i=0; i < w->tsp.n; i++ )
			XFillRectangle( XtDisplay( w ), XtWindow( w ), w->tsp.townGC,
				w->tsp.xpos[i], w->tsp.ypos[i], 2, 2 );
	}
}

void	XtDrawTour( w )
TspWidget	w;
{
	int	i;

	if ( w->tsp.tour_stored == FALSE ) return;

	if ( w->tsp.old_tour_stored == FALSE ) {
	/* Draw the first tour */
		for (i=0; i < w->tsp.n; i++ ) {
			XDrawLine( XtDisplay( w ), XtWindow( w ), w->tsp.new_tour_GC,
				w->tsp.xpos[ i ], w->tsp.ypos[ i ],
				w->tsp.xpos[ w->tsp.tour[i].t1 ], w->tsp.ypos[ w->tsp.tour[i].t1 ] );
		}
	} else {
		for (i=0; i < w->tsp.n; i++ ) {
			if ( (w->tsp.tour[i].t1 != w->tsp.old_tour[i].t1)
				&& (w->tsp.tour[i].t1 != w->tsp.old_tour[i].t2) ) {
			/* the link between i -> tour[i].t1 is new */
			XDrawLine( XtDisplay( w ), XtWindow( w ), w->tsp.new_tour_GC,
				w->tsp.xpos[ i ], w->tsp.ypos[ i ],
				w->tsp.xpos[ w->tsp.tour[i].t1 ], w->tsp.ypos[ w->tsp.tour[i].t1 ] );
			} else {
			/* the link between i -> tour[i].t1 is old */
			XDrawLine( XtDisplay( w ), XtWindow( w ), w->tsp.old_tour_GC,
				w->tsp.xpos[ i ], w->tsp.ypos[ i ],
				w->tsp.xpos[ w->tsp.tour[i].t1 ], w->tsp.ypos[ w->tsp.tour[i].t1 ] );
			}
		}
	}
}

void	XtDeleteTour( w )
TspWidget	w;
{
	int	i;

	if ( w->tsp.old_tour_stored == FALSE ) return;

	for (i=0; i < w->tsp.n; i++ ) {
		XDrawLine( XtDisplay( w ), XtWindow( w ), w->tsp.clGC,
				w->tsp.xpos[ i ], w->tsp.ypos[ i ],
				w->tsp.xpos[ w->tsp.old_tour[i].t1 ], w->tsp.ypos[ w->tsp.old_tour[i].t1 ] );
	}
}

void	XtClean( w )
TspWidget	w;
{
	XClearWindow( XtDisplay( w ), XtWindow( w ) );
}

void	XtDraw( w, t )
TspWidget	w;
int		*t;			/* Tour description */
{
	Neighbor	*p;
	int	n = w->tsp.n;
	int	i;

	/* compute new neighborhood */

	if ( w->tsp.tour_stored == FALSE ) {
		w->tsp.tour_stored = TRUE;
		w->tsp.old_tour_stored = FALSE;
	} else {
		w->tsp.old_tour_stored = TRUE;
	}
	for ( i=0; i < n; i++ ) {
		w->tsp.tour_buf[t[i]].t1 = t[(i+n-1) % n];
		w->tsp.tour_buf[t[i]].t2 = t[(i+1) % n];
	}

	/* delete old links between towns */

	if ( w->tsp.old_tour_stored == TRUE )
		for ( i=0; i < n; i++ ) {
			if ( (w->tsp.tour[i].t1 != w->tsp.tour_buf[i].t1)
			  && (w->tsp.tour[i].t1 != w->tsp.tour_buf[i].t2) ) {
				XDrawLine( XtDisplay( w ), XtWindow( w ), w->tsp.clGC,
					w->tsp.xpos[ i ], w->tsp.ypos[ i ],
					w->tsp.xpos[ w->tsp.tour[i].t1 ], w->tsp.ypos[ w->tsp.tour[i].t1 ] );
			}
		}

	/* draw new links between towns */

	if ( w->tsp.old_tour_stored == FALSE ) {
	/* Draw the first tour */
		for ( i=0; i < n; i++ )
			XDrawLine( XtDisplay( w ), XtWindow( w ), w->tsp.new_tour_GC,
				w->tsp.xpos[ i ], w->tsp.ypos[ i ],
				w->tsp.xpos[ w->tsp.tour_buf[i].t1 ], w->tsp.ypos[ w->tsp.tour_buf[i].t1 ] );
	} else {
		for ( i=0; i < n; i++ )
			if ( (w->tsp.tour_buf[i].t1 != w->tsp.tour[i].t1)
			  && (w->tsp.tour_buf[i].t1 != w->tsp.tour[i].t2) )
				XDrawLine( XtDisplay( w ), XtWindow( w ), w->tsp.new_tour_GC,
					w->tsp.xpos[ i ], w->tsp.ypos[ i ],
					w->tsp.xpos[ w->tsp.tour_buf[i].t1 ], w->tsp.ypos[ w->tsp.tour_buf[i].t1 ] );
			else
				XDrawLine( XtDisplay( w ), XtWindow( w ), w->tsp.old_tour_GC,
					w->tsp.xpos[ i ], w->tsp.ypos[ i ],
					w->tsp.xpos[ w->tsp.tour_buf[i].t1 ], w->tsp.ypos[ w->tsp.tour_buf[i].t1 ] );
	}

	p = w->tsp.old_tour;
	w->tsp.old_tour = w->tsp.tour;		/* set new old tour */
	w->tsp.tour = w->tsp.tour_buf;		/* set new new tour */
	w->tsp.tour_buf = p;

	XtDrawTowns( w );
}

