/******************************************************************************
*
*	link interface using parsytecs 'link.o' routines.
*	17 January 1990 mc Parsytec
*
*  File : "linkio.c"
*
*
*  Copyright PARSYTEC / INMOS Limited, 1990.
*
******************************************************************************/

#define TRUE  1
#define FALSE 0

#define	BYTE_TIMEOUT	50	/* in units of 0.1 seconds */

#include <stdio.h>

/* -- link driver variables */
extern	int LinkId;

/* -- link driver procedures */
/* link_init 
 */
int link_init ( tnumber )
int	tnumber;
{	int ret;
	char link_name[10];	/* decimal string representation of tnumber */

	/* convert 'tnumber' into decimal string */
	sprintf (link_name, "%d", tnumber);

	/* try to open the link 'tnumber' */
	if ( (ret = OpenLink (link_name)) < 0) {
		/* failed to open */
		/* poor error handling now */
		fprintf (stderr, "\nlink_init: Failed to open %d, because %s\n",			tnumber, lerror (ret));
		return (FALSE);
	} else {
		/* opened ok, store returned LinkId */
		LinkId = ret;
		return (TRUE);
	}
}

/* -- link_in_test */
int link_in_test ()
/* Return TRUE if a byte is available on the link, FALSE otherwise. */
{
	    return( TestRead (LinkId) );
}

/* -- byte_to_link */
void byte_to_link (ch)
unsigned char	ch;
/*  write a byte to the linkadaptor */
{	int ret;
	ret = WriteLink (LinkId, &ch, 1, 0);	/* timeout = eternaty */
	/* future expansion: error handling */
}


/* -- byte_to_link_t */
int byte_to_link_t (ch)
unsigned char ch;
/* Write a byte to the link adaptor within a specified time, or timeout.
   Return TRUE if the byte was NOT written, FALSE if the byte was written. */
{	int ret;
	ret = WriteLink (LinkId, &ch, 1, BYTE_TIMEOUT);
	if ( ret == 1 ) {
		/* fine */
		return (FALSE);
	} else {
		/* timeout (ret==0) or error. see note below */
		return (TRUE);
	}
	/* future expansion: error handling */
}


/* -- byte_from_link */
int byte_from_link ()
/*  read a byte from the linkadaptor. No sign extension! */
{	int ret;
	unsigned char ch;

	ret = ReadLink (LinkId, &ch, 1, 0);	/* timeout = eternaty */
	/* future expansion: error handling */
	if (ret==1) {
		return ( (unsigned int) ch );
	}
}

/* -- reset_root */
void reset_root ()
/*  reset the root transputer */
{
	ResetLink (LinkId);
	/* for future: include error handling */
}

/* -- link word and slice operations */
/* -- word_from_link */
int word_from_link ()
/* Read a word from the link. */
/* This routine is host- and transputer- word-length dependent as well as
 * byte-ordering dependend.
 * The transputer sends the bytes low --> high.
 * Here we assume a host int length of four bytes, and a transputer int length
 * of four bytes.
*/
{   register unsigned int result, i;

    result = 0;
    for (i=0; i<32; i+=8)	/* receive bytes low --> high */
	result |= ( ((unsigned int)byte_from_link () ) << i);
    return (result);
}

/* -- word_to_link */
void word_to_link (w)
register int w;
/* Write a word to the link, least significant byte first. */
/* This routine is host- and transputer- word-length dependent as well as
 * byte-ordering dependend.
 * The transputer sends the bytes low --> high.
 * Here we assume a host int length of four bytes, and a transputer int length
 * of four bytes.
*/
{	register short i;

	for (i=0; i<4; i++) {
		byte_to_link ( w & 0xff );
		w >>= 8;
	}
}

/* -- word_to_link_t */
int word_to_link_t (w)
register int w;
/* Try to write a word to the link in time, least significant byte first. */
/* Return TRUE on failure to write the hole word, else FALSE (==> O.K.).  */
/* This routine is host- and transputer- word-length dependent as well as
 * byte-ordering dependend.
 * The transputer sends the bytes low --> high.
 * Here we assume a host int length of four bytes, and a transputer int length
 * of four bytes.
*/
{	register short i; 

	for (i=0; i<4; i++) {
		if (byte_to_link_t ( w & 0xff ) )
			/* Timeout. */
			return (TRUE);
		w >>= 8;
	}
	return (FALSE);		/* word written in time */
}


/* -- long_word_to_link */
void long_word_to_link (w)
long w;
/* Write a long word to the link. */
/* This routine is host- and transputer- word-length dependent as well as
 * byte-ordering dependend.
 * The transputer sends the bytes low --> high.
 * Here we assume a host int length of four bytes, and a transputer int length
 * of four bytes.
*/
{
	word_to_link (w);
}

/* -- long_word_from_link */
long long_word_from_link ()
/* Read a long word from the link. */
/* This routine is host- and transputer- word-length dependent as well as
 * byte-ordering dependend.
 * The transputer sends the bytes low --> high.
 * Here we assume a host int length of four bytes, and a transputer int length
 * of four bytes.
*/

{
    return ( word_from_link () ); 
}

/****************************************************************************/
/*                       Extensions                                         */
/****************************************************************************/

getbuf(buf, nbytes)
char *buf;
unsigned int nbytes; 
{   int ret;

    ret = ReadLink (LinkId, buf, nbytes, 0);
    if (ret < 0) {
	fprintf (stderr, "\ngetbuf: error %s\n", lerror (ret));
    }
}

putbuf(buf, nbytes)
char *buf;
unsigned int nbytes; 
{   int ret;
    int time_out;

    time_out = nbytes/50 + 50;	/* 5 sec + 2 sec/KByte */
    ret = WriteLink (LinkId, buf, nbytes, time_out);
    if (ret < 0) {
	fprintf (stderr, "\nputbuf: error %s\n", lerror (ret));
    } else if (ret != nbytes) {
		/* timed out */
		fprintf (stderr, "\nputbuf TIMEOUT after %d bytes, exiting\n",
			ret);
		exit (1);
    }
}


char *getstring() {
int size,i, ret;
char *str,*stst;
    int time_out;

    size = word_from_link() ;
    time_out = size/50 + 50;	/* 5 sec + 2 sec/KByte */
    stst = (char*)malloc (size + 1) ;
    str = stst ;
    ret = ReadLink (LinkId, stst, size, time_out);
    if (ret >= 0) {
	str += ret;
    } else {
	fprintf (stderr, "\ngetstring some error");
	/* poor error handling now */
    }
    *str=0;                                 /* make this a proper 'c' string */
    return (stst) ;
}

putstring(str) 
char *str; 
{
int size,i;
    int time_out, ret;

    size = strlen(str) + 1;
    time_out = size/50 + 50;	/* 5 sec + 2 sec/KByte */
    word_to_link_t(size);
    ret = WriteLink (LinkId, str, size, time_out);
    if (ret < 0) {
	fprintf (stderr, "\nputstring some error");
	/* poor error handling now */
    }
}
